"""File with weigthed regression model algorithm."""
# Third Party Library
import numpy as np

from pamm.algorithms.regression.regression_model import RegressionAlgorithm


class ModifyRegressionAlgorithm(RegressionAlgorithm):
    def _add_bias(self, x: np.ndarray, q: np.ndarray, b_apr: np.ndarray) -> tuple[np.ndarray]:
        if self._fit_bias:
            return (np.hstack([np.ones((x.shape[0], 1)), x]), np.diag(np.hstack([0, np.diag(q)])), np.vstack([0, b_apr]))
        return x, q, b_apr

    def fit(self, x: np.ndarray, y: np.ndarray, w: np.ndarray, q: np.ndarray, b_apr: np.ndarray) -> None:
        """Fit model.

        Args:
            x (np.ndarray): matrix of regressors
            y (np.ndarray): matrix a teacher
            w (np.ndarray): matrix of weights for data
            q (np.ndarray): matrix of weights for beta
            b_apr (np.ndarray): knowns beta coefficients
        """

        x, q, b_apr = self._add_bias(x, q, b_apr)
        tmp = b_apr + (np.linalg.inv(q + x.T @ w @ x) @ x.T @ w) @ (y - x @ b_apr)
        self._beta = tmp[int(self._fit_bias) :]
        self._bias = tmp[0][0] * int(self._fit_bias)



from pamm.metricks import numeric as nm


class TestNumericMetrick:
    
    def test_mse(self, get_vectors):
        solution = nm.mse(major_ts = get_vectors['major_ts'], minor_ts = get_vectors['minor_ts'])
        assert solution == (get_vectors['shift']**2)
        
        
    def test_rmse(self, get_vectors):
        solution = nm.rmse(major_ts = get_vectors['major_ts'], minor_ts = get_vectors['minor_ts'])
        assert solution == get_vectors['shift']
        
        
    def test_mae(self, get_vectors):
        solution = nm.mae(major_ts = get_vectors['major_ts'], minor_ts = get_vectors['minor_ts'])
        assert solution == get_vectors['shift']
        
    
    def test_mspe(self, get_vectors):
        solution = nm.mspe(major_ts = get_vectors['major_ts'], minor_ts = get_vectors['minor_ts'])
        assert solution > get_vectors['shift']
        
    
    def test_mape(self, get_vectors):
        solution = nm.mape(major_ts = get_vectors['major_ts'], minor_ts = get_vectors['minor_ts'])
        assert solution > get_vectors['shift']
        
        
    def test_smape(self, get_vectors):
        solution = nm.smape(major_ts = get_vectors['major_ts'], minor_ts = get_vectors['minor_ts'])
        assert solution > get_vectors['shift']
        
        
    def test_mre(self, get_vectors):
        solution = nm.mre(major_ts = get_vectors['major_ts'], minor_ts = get_vectors['minor_ts'])
        assert solution > get_vectors['shift']
        
        
    def test_rmsle(self, get_vectors):
        solution = nm.rmsle(major_ts = get_vectors['major_ts'], minor_ts = get_vectors['minor_ts'])
        assert solution < get_vectors['shift']
        
        
    def test_corr(self, get_vectors):
        solution = nm.corr(major_ts = get_vectors['major_ts'], minor_ts = get_vectors['minor_ts'])
        assert solution > .98
        
        
    def test_r2(self, get_vectors):
        solution = nm.r2(major_ts = get_vectors['major_ts'], minor_ts = get_vectors['minor_ts'])
        assert solution < get_vectors['shift']
        
        
    def test_adjusted_r2(self, get_vectors):
        solution = nm.adjusted_r2(major_ts = get_vectors['major_ts'], minor_ts = get_vectors['minor_ts'], count_features= get_vectors['shift'])
        assert solution < get_vectors['shift']


# Standard Library
from abc import ABC, abstractmethod


class AbstractBaseModelAlgorithm(ABC):
    @abstractmethod
    def fit(self):
        """Get fitting of model."""

    @abstractmethod
    def predict(self):
        """Get predict of model."""


class AbstractRegressionModelAlgorithm(ABC):
    @property
    @abstractmethod
    def beta(self):
        """Get coefficients of model."""

    @property
    @abstractmethod
    def bias(self):
        """Get bias of model."""


class AbstractTreeModelAlgorithm(ABC):
    """Abstruct for Tree"""


class AbstractNeyroModelAlgorithm(ABC):
    """Abstruct for Neyro"""



"""File with regression model algorithm on two-stage least sqares."""
# Third Party Library
import numpy as np

from pamm.algorithms.regression.regression_model import RegressionAlgorithm


class StachosticGradientRegressionAlgorithm(RegressionAlgorithm):
    def fit(self, x: np.ndarray, y: np.ndarray, eta: float = 0.1, n_epochs: int = 10, t0: int = 1, t1: int = 10) -> np.ndarray:
        x = self._add_bias(x)
        m, n = x.shape
        theta = np.random.randn(n, 1)
        for epoch in range(n_epochs):
            for i in range(m):
                random_index = np.random.randint(m)
                xi = x[random_index : random_index + 1]
                yi = y[random_index : random_index + 1]
                gradients = 2 * xi.T @ (xi @ theta - yi)
                eta = t0 / ((epoch * m + i) + t1)
                theta = theta - eta * gradients

        self._beta = theta[int(self._fit_bias) :]
        self._bias = theta[0][0] * int(self._fit_bias)



"""File with quad prog algorithm."""
# Third Party Library
import cvxopt as quad_prog_cvx
import numpy as np

from pamm.algorithms.regression.regression_model import RegressionAlgorithm


class QPRegressionAlgorithm(RegressionAlgorithm):

    """Class for calculating linear model coefficients using MRLS(weighted Least Squares) + Quadprog."""

    def _add_bias(self, x: np.ndarray, min: np.ndarray, max: np.ndarray) -> tuple[np.ndarray]:
        if self._fit_bias:
            return (np.hstack([np.ones((x.shape[0], 1)), x]), np.vstack([-100000, min]), np.vstack([100000, max]))
        return x, min, max

    def fit(self, x: np.ndarray, y: np.ndarray, r: np.array, beta_min: np.ndarray, beta_max: np.ndarray) -> None:
        """The method of calculating coefficients is the RLS method using quadprog.

        Args:
            x_matrix (np.ndarray): matrix of regressors
            y_vector (np.ndarray): teacher vector
            r_matrix (np.ndarray): data trust matrix
            beta_min (np.atleast_2d): tuple with minimum values of regressors
            beta_max (np.atleast_2d): tuple with maximum values of regressors
        """

        x, beta_min, beta_max = self._add_bias(x, beta_min, beta_max)

        _, count_features = x.shape

        h = 2 * (x.T @ r @ x)
        f = (-2 * (y.T @ r @ x)).T
        a = np.concatenate((np.identity(count_features), -np.identity(count_features)), axis=0)
        b = np.concatenate((beta_max, -beta_min), axis=0)

        p = quad_prog_cvx.matrix(h.astype(np.double))
        q = quad_prog_cvx.matrix(f.astype(np.double))
        g = quad_prog_cvx.matrix(a.astype(np.double))
        h = quad_prog_cvx.matrix(b.astype(np.double))

        quad_prog_cvx.solvers.options['kktsolver'] = 'ldl'
        quad_prog_cvx.solvers.options['feastol'] = 1e-16
        quad_prog_cvx.solvers.options['maxiters'] = 1000
        res_quadprog = quad_prog_cvx.solvers.qp(P=p, q=q, G=g, h=h)["x"].T
        tmp = np.atleast_2d(res_quadprog)[0].reshape(count_features, 1)

        self._beta = tmp[int(self._fit_bias) :]
        self._bias = tmp[0][0] * int(self._fit_bias)



import numpy as np


def bgd(x: np.ndarray, y: np.ndarray, eta: float = 0.1, max_iter: int = 1000) -> np.ndarray:
    m, n = x.shape
    theta = np.random.randn(n, 1)
    for iteration in range(max_iter):
        gradients = 2 / m * x.T @ (x @ theta - y)
        theta = theta - eta * gradients
    return theta


def sgd(x: np.ndarray, y: np.ndarray, eta: float = 0.1, n_epochs: int = 10, t0: int = 1, t1: int = 10) -> np.ndarray:
    m, n = x.shape
    theta = np.random.randn(n, 1)
    for epoch in range(n_epochs):
        for i in range(m):
            random_index = np.random.randint(m)
            xi = x[random_index : random_index + 1]
            yi = y[random_index : random_index + 1]
            gradients = 2 * xi.T @ (xi @ theta - yi)
            eta = t0 / ((epoch * m + i) + t1)
            theta = theta - eta * gradients
    return theta



import numpy as np


def set_delay(ts: np.ndarray, delay: int) -> np.ndarray:
    """Add delation for time-series.

    Args:
        ts (np.ndarray): time-seriess.
        delay (int): delay in int format.

    Returns:
        np.ndarray: filtered time-seriess.
    """
    data = ts.copy()
    data[:, 0] = data[:, 0] + delay
    return data


def drop_date(ts: np.ndarray, start: int, end: int) -> np.ndarray:
    """Dropped time interval from time-series.

    Args:
        ts (np.ndarray): time-seriess.
        start (int): start date for drop.
        end (int): end date for drop.

    Returns:
        np.ndarray: filtered time-seriess.
    """
    values = ts[:, 0]
    mask = (values < end), (values > start)
    return ts[np.logical_and(*mask)]



"""Conftest for genering test-data"""

import pytest
import random
import numpy as np


@pytest.fixture(scope='module')
def get_ts():
    size_major, size_minor = random.randint(1,437), random.randint(1000-7,27182)
    major_ts = np.random.rand(size_major, 2)
    minor_ts = np.random.rand(size_minor, 2)
    
    return {'major_ts': major_ts, 'minor_ts': minor_ts}


@pytest.fixture(scope='module')
def get_vectors():
    num = random.randint(1,437)
    shift = random.randint(3,22)
    major_ts = np.random.rand(num, 1)
    minor_ts = shift + major_ts
    return {'major_ts': major_ts, 'minor_ts': minor_ts, 'shift': shift}


@pytest.fixture(scope = "module")
def get_data(count_beta: int =  5):
    beta = np.random.rand(count_beta, 1)
    x = np.random.rand(count_beta * 20, count_beta)
    z = np.random.rand(count_beta * 20, count_beta + 2)
    w = np.eye(count_beta * 20)
    beta_apr = np.ones(count_beta).reshape((count_beta,1))
    q = np.eye(count_beta)
    beta_min = np.ones(count_beta).reshape((count_beta,1)) * -1
    beta_min[0,0] = 0.0000000009
    beta_max = np.ones(count_beta).reshape((count_beta,1)) * 5
    beta_max[0,0] = 0.09
    y = x @ beta
    return {'x': x, 'z': z, 'w': w, 'y': y, 'q': q, 'beta_apr': beta_apr, 'beta': beta, 'beta_max': beta_max, 'beta_min': beta_min}


from pamm.algorithms.regression import(RegressionAlgorithm,
  TwoStageLeastSquaresAlgorithm,
  JackKnifeRegressionAlgorithm,
  IterationRegressionAlgorithm,
  RidgeRegressionAlgorithm,
  LassoRegressionAlgorithm,
  ElasticNetRegressionAlgorithm,
  WeigthedRegressionAlgorithm,
  WeightedTwoStageLeastSquaresAlgorithm,
  ModifyRegressionAlgorithm,
  QPRegressionAlgorithm,
  BatchGradientRegressionAlgorithm,
  StachosticGradientRegressionAlgorithm)


class TestBaseMethods:
    
    def test_predict(self, get_data):
        model = RegressionAlgorithm(fit_bias = True)
        model.fit(x = get_data['x'], y = get_data['y'])
        assert model.predict(get_data['x']).any() > 0
        
        
    def test_set_beta(self):
        model = RegressionAlgorithm(fit_bias = True)
        model.beta = (1,2,3,4)
        assert model.beta.shape == (4,1)
        
        
    def test_bias(self):
        model = RegressionAlgorithm(fit_bias=True)
        model.bias = 1
        assert model.bias == 1
        
        
    def test_fit_bias_true(self, get_data):
        model = RegressionAlgorithm(fit_bias = True)
        model.fit(x = get_data['x'], y = get_data['y'])
        assert model.bias != 0
        
        
    def test_fit_bias_false(self, get_data):
        model = RegressionAlgorithm(fit_bias = False)
        model.fit(x = get_data['x'], y = get_data['y'])
        assert model.bias == 0
            
class TestNoParametrizeAlgorithm:
    
    def test_regression_algorithm(self, get_data):
        model = RegressionAlgorithm(fit_bias = True)
        model.fit(x = get_data['x'], y = get_data['y'])
        assert model.beta.shape == get_data['beta'].shape
        
        
    def test_tsls_algoritm(self, get_data):
        model = TwoStageLeastSquaresAlgorithm(fit_bias = True)
        model.fit(x = get_data['x'], z = get_data['z'], y = get_data['y'])
        assert model.beta.shape == get_data['beta'].shape
        
        
    def test_jackknife_algoritm(self, get_data):
        model = JackKnifeRegressionAlgorithm(fit_bias = True)
        model.fit(x = get_data['x'], y = get_data['y'])
        assert model.beta.shape == get_data['beta'].shape
     
         
    def test_iterations_algoritm(self, get_data):
        model = IterationRegressionAlgorithm(fit_bias = True)
        model.fit(x = get_data['x'], y = get_data['y'])
        assert model.beta.shape == get_data['beta'].shape
        
class TestRegularizationAlgorithm:

    def test_ridge_algorithm(self, get_data):
        model = RidgeRegressionAlgorithm(fit_bias = True)
        model.fit(x = get_data['x'], y = get_data['y'])
        assert model.beta.shape == get_data['beta'].shape
        
        
    def test_lasso_algorithm(self, get_data):
        model = LassoRegressionAlgorithm(fit_bias = True)
        model.fit(x = get_data['x'], y = get_data['y'])
        assert model.beta.shape == get_data['beta'].shape
        
        
    def test_elastic_net_algorithm(self, get_data):
        model = ElasticNetRegressionAlgorithm(fit_bias = True)
        model.fit(x = get_data['x'], y = get_data['y'])
        assert model.beta.shape == get_data['beta'].shape
        
class TestWeightedAlgorithm:
    
    def test_wls_algorithm(self, get_data):
        model = WeigthedRegressionAlgorithm(fit_bias = True)
        model.fit(x = get_data['x'], w=get_data['w'], y = get_data['y'])
        assert model.beta.shape == get_data['beta'].shape
        
        
    def test_wtsls_algorithm(self, get_data):
        model = WeightedTwoStageLeastSquaresAlgorithm(fit_bias = True)
        model.fit(x = get_data['x'], z = get_data['z'], w=get_data['w'], y = get_data['y'])
        assert model.beta.shape == get_data['beta'].shape
    
    
    def test_modify_algorithm(self, get_data):
        model = ModifyRegressionAlgorithm(fit_bias = True)
        model.fit(x = get_data['x'], q = get_data['q'], b_apr=get_data['beta_apr'], w=get_data['w'], y = get_data['y'])
        assert model.beta.shape == get_data['beta'].shape
        
        
    def test_quad_prog_algorithm(self, get_data):
        model = QPRegressionAlgorithm(fit_bias = True)
        model.fit(x = get_data['x'], beta_min = get_data['beta_min'], beta_max = get_data['beta_max'], r=get_data['w'], y = get_data['y'])
        assert model.beta.shape == get_data['beta'].shape
    
class TestGradientDeskAlgorithm:
    
    def test_batch_gradient_algorithm(self, get_data):
        model = BatchGradientRegressionAlgorithm(fit_bias = True)
        model.fit(x = get_data['x'], y = get_data['y'])
        assert model.beta.shape == get_data['beta'].shape
        
    
    def test_stachostic_gradient_algorithm(self, get_data):
        model = StachosticGradientRegressionAlgorithm(fit_bias = True)
        model.fit(x = get_data['x'], y = get_data['y'])
        assert model.beta.shape == get_data['beta'].shape
    


from pamm.algorithms.weighing.weights_data_algorithms import inversely, equal, time_booster
from pamm.algorithms.weighing.weigts_beta_algorithms import max_truth


import random

class TestBetaWeights:
    
    def test_max_truth(self, get_data):
        res = max_truth(get_data['beta'])
        assert get_data['beta'].shape[0] == res.shape[random.randint(0,1)]

class TestDataWeights:
    
    def test_inversely(self, get_ts):
        res = inversely(get_ts['minor_ts'])
        assert get_ts['minor_ts'].shape[0] == res.shape[random.randint(0,1)]
        
    
    def test_equal(self, get_ts):
        res = equal(get_ts['minor_ts'])
        size = get_ts['minor_ts'].shape[0]
        rnd_number = random.randint(0,size)
        assert size == res.shape[random.randint(0,1)]
        assert res[rnd_number,rnd_number] == 1
        
    
    def test_time_booster(self, get_ts):
        res = time_booster(get_ts['minor_ts'])
        assert get_ts['minor_ts'].shape[0] == res.shape[random.randint(0,1)]
          


import numpy as np


def peccentel_edge(ts: np.ndarray, percentile: float) -> np.ndarray:
    """Delete all values bihind percentel for two edges.

    Args:
        ts (np.ndarray): time-series.
        percentile (float): percentile data.

    Returns:
        np.ndarray: filtered time-series.
    """
    values = ts[:, 1]
    up_percentile = np.percentile(values, percentile)
    low_percentile = np.percentile(values, 100 - percentile)
    mask = (values < up_percentile), (values > low_percentile)
    return ts[np.logical_and(*mask)]


def percentel(ts: np.ndarray, percentile: float) -> np.ndarray:
    """Delete all values bihind percentel.

    Args:
        ts (np.ndarray): time-series.
        percentile (float): percentile data.

    Returns:
        np.ndarray: filtered time-series.
    """
    values = ts[:, 1]
    percent = np.percentile(values, percentile)
    mask = values < percent
    return ts[np.logical_and(*mask)]


def power(ts: np.ndarray, pow: float = 2) -> np.ndarray:
    """Delete all values bihind percentel.

    Args:
        ts (np.ndarray): time-series.
        percentile (float): percentile data.

    Returns:
        np.ndarray: filtered time-series.
    """
    ts[:, 1] = ts[:, 1] ** pow
    return ts


def inside(ts: np.ndarray, min: int | None = None, max: int | None = None) -> np.ndarray:
    """Return values only inside interval.

    Args:
        ts (np.ndarray): time-seriess.
        min (int): start date for drop.
        max (int): end date for drop.

    Returns:
        np.ndarray: filtered time-seriess.
    """
    min = min if min else -np.inf
    max = max if max else np.inf
    values = ts[:, 0]
    mask = (values > min), (values < max)
    return ts[np.logical_and(*mask)]


def moving_average(ts: np.ndarray, window: int) -> np.ndarray:
    """Return values only inside interval.

    Args:
        ts (np.ndarray): time-seriess.
        window (int): sliding window.

    Returns:
        np.ndarray: filtered time-seriess.
    """
    times = np.convolve(ts[:, 0], np.ones(window), "valid") / window
    values = np.convolve(ts[:, 1], np.ones(window), "valid") / window
    #FIXME np.column_stack не нужно транспонировать
    return np.vstack([times, values]).T
